# Copyright 2016 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Dummy share driver for testing Manila APIs and other interfaces.

This driver simulates support of:
- Both available driver modes: DHSS=True/False
- NFS and CIFS protocols
- IP access for NFS shares and USER access for CIFS shares
- CIFS shares in DHSS=True driver mode
- Creation and deletion of share snapshots
- Share replication (readable)
- Share migration
- Consistency groups
- Resize of a share (extend/shrink)

"""

import functools
import time

from oslo_config import cfg
from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import timeutils

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import configuration
from manila.share import driver
from manila.share.manager import share_manager_opts  # noqa
from manila.share import utils as share_utils

LOG = log.getLogger(__name__)


dummy_opts = [
    cfg.FloatOpt(
        "dummy_driver_default_driver_method_delay",
        help="Defines default time delay in seconds for each dummy driver "
             "method. To redefine some specific method delay use other "
             "'dummy_driver_driver_methods_delays' config opt. Optional.",
        default=2.0,
        min=0,
    ),
    cfg.DictOpt(
        "dummy_driver_driver_methods_delays",
        help="It is dictionary-like config option, that consists of "
             "driver method names as keys and integer/float values that are "
             "time delay in seconds. Optional.",
        default={
            "ensure_share": "1.05",
            "create_share": "3.98",
            "get_pool": "0.5",
            "do_setup": "0.05",

            "_get_pools_info": "0.1",
            "_update_share_stats": "0.3",

            "create_replica": "3.99",
            "delete_replica": "2.98",
            "promote_replica": "0.75",
            "update_replica_state": "0.85",
            "create_replicated_snapshot": "4.15",
            "delete_replicated_snapshot": "3.16",
            "update_replicated_snapshot": "1.17",

            "migration_start": 1.01,
            "migration_continue": 1.02,  # it will be called 2 times
            "migration_complete": 1.03,
            "migration_cancel": 1.04,
            "migration_get_progress": 1.05,
            "migration_check_compatibility": 0.05,
        },
    ),
]

CONF = cfg.CONF


def slow_me_down(f):

    @functools.wraps(f)
    def wrapped_func(self, *args, **kwargs):
        sleep_time = self.configuration.safe_get(
            "dummy_driver_driver_methods_delays").get(
                f.__name__,
                self.configuration.safe_get(
                    "dummy_driver_default_driver_method_delay")
            )
        time.sleep(float(sleep_time))
        return f(self, *args, **kwargs)

    return wrapped_func


def get_backend_configuration(backend_name):
    config_stanzas = CONF.list_all_sections()
    if backend_name not in config_stanzas:
        msg = _("Could not find backend stanza %(backend_name)s in "
                "configuration which is required for share replication and "
                "migration. Available stanzas are %(stanzas)s")
        params = {
            "stanzas": config_stanzas,
            "backend_name": backend_name,
        }
        raise exception.BadConfigurationException(reason=msg % params)

    config = configuration.Configuration(
        driver.share_opts, config_group=backend_name)
    config.append_config_values(dummy_opts)
    config.append_config_values(share_manager_opts)
    config.append_config_values(driver.ssh_opts)

    return config


class DummyDriver(driver.ShareDriver):
    """Dummy share driver that implements all share driver interfaces."""

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        super(DummyDriver, self).__init__(
            [False, True], *args, config_opts=[dummy_opts], **kwargs)
        self._verify_configuration()
        self.private_storage = kwargs.get('private_storage')
        self.backend_name = self.configuration.safe_get(
            "share_backend_name") or "DummyDriver"
        self.migration_progress = {}
        self.security_service_update_support = True
        self.network_allocation_update_support = True

    def _verify_configuration(self):
        allowed_driver_methods = [m for m in dir(self) if m[0] != '_']
        allowed_driver_methods.extend([
            "_setup_server",
            "_teardown_server",
            "_get_pools_info",
            "_update_share_stats",
        ])
        disallowed_driver_methods = (
            "get_admin_network_allocations_number",
            "get_network_allocations_number",
            "get_share_server_pools",
        )
        for k, v in self.configuration.safe_get(
                "dummy_driver_driver_methods_delays").items():
            if k not in allowed_driver_methods:
                raise exception.BadConfigurationException(reason=(
                    "Dummy driver does not have '%s' method." % k
                ))
            elif k in disallowed_driver_methods:
                raise exception.BadConfigurationException(reason=(
                    "Method '%s' does not support delaying." % k
                ))
            try:
                float(v)
            except (TypeError, ValueError):
                raise exception.BadConfigurationException(reason=(
                    "Wrong value (%(v)s) for '%(k)s' dummy driver method time "
                    "delay is set in 'dummy_driver_driver_methods_delays' "
                    "config option." % {"k": k, "v": v}
                ))

    def _get_share_name(self, share):
        return "share_%(s_id)s_%(si_id)s" % {
            "s_id": share["share_id"].replace("-", "_"),
            "si_id": share["id"].replace("-", "_")}

    def _get_snapshot_name(self, snapshot):
        return "snapshot_%(s_id)s_%(si_id)s" % {
            "s_id": snapshot["snapshot_id"].replace("-", "_"),
            "si_id": snapshot["id"].replace("-", "_")}

    def _get_export(self, mountpoint, ip, is_admin_only, preferred):
        return {
            "path": "%(ip)s:%(mp)s" % {"ip": ip, "mp": mountpoint},
            "metadata": {
                "preferred": preferred,
            },
            "is_admin_only": is_admin_only,
        }

    def _get_subnet_allocations_from_backend_details(self, backend_details):
        """Reads subnet_allocations info from backend details"""
        # NOTE(sfernand): Ensure backward compatibility for share servers
        # created prior to the addition of support to multiple subnets per AZ,
        # by read ip information using the old format in case
        # subnet_allocations does not exist.
        if 'subnet_allocations' in backend_details:
            subnet_allocations = jsonutils.loads(
                backend_details['subnet_allocations'])
        else:
            subnet_allocations = [{
                'primary_public_ip':
                    backend_details['primary_public_ip'],
                'secondary_public_ip':
                    backend_details['secondary_public_ip']
            }]
        return subnet_allocations

    def _generate_export_locations(self, mountpoint, share_server=None):
        if share_server:
            backend_details = share_server['backend_details']
            subnet_allocations = (
                self._get_subnet_allocations_from_backend_details(
                    backend_details))
            service_ip = backend_details["service_ip"]
        else:
            subnet_allocations = [{
                "primary_public_ip": "10.0.0.10",
                "secondary_public_ip": "10.0.0.20",
            }]
            service_ip = "11.0.0.11"

        export_locations = [
            self._get_export(mountpoint, service_ip, True, False)]
        for subnet_allocation in subnet_allocations:
            export_locations.append(
                self._get_export(
                    mountpoint, subnet_allocation["primary_public_ip"],
                    False, True))
            export_locations.append(
                self._get_export(
                    mountpoint, subnet_allocation["secondary_public_ip"],
                    False, False))

        return export_locations

    def _create_share(self, share, share_server=None):
        share_proto = share["share_proto"]
        if share_proto not in ("NFS", "CIFS"):
            msg = _("Unsupported share protocol provided - %s.") % share_proto
            raise exception.InvalidShareAccess(reason=msg)

        share_name = self._get_share_name(share)
        mountpoint = "/path/to/fake/share/%s" % share_name
        self.private_storage.update(
            share["id"], {
                "fake_provider_share_name": share_name,
                "fake_provider_location": mountpoint,
            }
        )
        return self._generate_export_locations(
            mountpoint, share_server=share_server)

    @slow_me_down
    def create_share(self, context, share, share_server=None):
        """Is called to create share."""
        return self._create_share(share, share_server=share_server)

    @slow_me_down
    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Is called to create share from snapshot."""
        export_locations = self._create_share(share, share_server=share_server)
        return {
            'export_locations': export_locations,
            'status': constants.STATUS_AVAILABLE
        }

    def _create_snapshot(self, snapshot, share_server=None):
        snapshot_name = self._get_snapshot_name(snapshot)
        mountpoint = "/path/to/fake/snapshot/%s" % snapshot_name
        self.private_storage.update(
            snapshot["id"], {
                "fake_provider_snapshot_name": snapshot_name,
                "fake_provider_location": mountpoint,
            }
        )
        return {
            'fake_key1': 'fake_value1',
            'fake_key2': 'fake_value2',
            'fake_key3': 'fake_value3',
            "provider_location": mountpoint,
            "export_locations": self._generate_export_locations(
                mountpoint, share_server=share_server)
        }

    @slow_me_down
    def create_snapshot(self, context, snapshot, share_server=None):
        """Is called to create snapshot."""
        return self._create_snapshot(snapshot, share_server)

    @slow_me_down
    def delete_share(self, context, share, share_server=None):
        """Is called to remove share."""
        self.private_storage.delete(share["id"])

    @slow_me_down
    def delete_snapshot(self, context, snapshot, share_server=None):
        """Is called to remove snapshot."""
        LOG.debug('Deleting snapshot with following data: %s', snapshot)
        self.private_storage.delete(snapshot["id"])

    @slow_me_down
    def get_pool(self, share):
        """Return pool name where the share resides on."""
        pool_name = share_utils.extract_host(share["host"], level="pool")
        return pool_name

    @slow_me_down
    def ensure_share(self, context, share, share_server=None):
        """Invoked to ensure that share is exported."""

    @slow_me_down
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share."""
        for rule in add_rules + access_rules:
            share_proto = share["share_proto"].lower()
            access_type = rule["access_type"].lower()
            if not (
                    (share_proto == "nfs" and access_type == "ip") or
                    (share_proto == "cifs" and access_type == "user")):
                msg = _("Unsupported '%(access_type)s' access type provided "
                        "for '%(share_proto)s' share protocol.") % {
                    "access_type": access_type, "share_proto": share_proto}
                raise exception.InvalidShareAccess(reason=msg)

    @slow_me_down
    def snapshot_update_access(self, context, snapshot, access_rules,
                               add_rules, delete_rules, share_server=None):
        """Update access rules for given snapshot."""
        self.update_access(context, snapshot['share'], access_rules,
                           add_rules, delete_rules, share_server)

    @slow_me_down
    def do_setup(self, context):
        """Any initialization the share driver does while starting."""

    @slow_me_down
    def manage_existing(self, share, driver_options):
        """Brings an existing share under Manila management."""
        new_export = share['export_location']
        old_share_id = self._get_share_id_from_export(new_export)
        old_export = self.private_storage.get(
            old_share_id, key='export_location')
        if old_export.split(":/")[-1] == new_export.split(":/")[-1]:
            result = {"size": 1, "export_locations": self._create_share(share)}
            self.private_storage.delete(old_share_id)
            return result
        else:
            msg = ("Invalid export specified, existing share %s"
                   " could not be found" % old_share_id)
            raise exception.ShareBackendException(msg=msg)

    @slow_me_down
    def manage_existing_with_server(
            self, share, driver_options, share_server=None):
        return self.manage_existing(share, driver_options)

    def _get_share_id_from_export(self, export_location):
        values = export_location.split('share_')
        if len(values) > 1:
            return values[1][37:].replace("_", "-")
        else:
            return export_location

    @slow_me_down
    def unmanage(self, share):
        """Removes the specified share from Manila management."""
        self.private_storage.update(
            share['id'], {'export_location': share['export_location']})

    @slow_me_down
    def unmanage_with_server(self, share, share_server=None):
        self.unmanage(share)

    @slow_me_down
    def manage_existing_snapshot_with_server(self, snapshot, driver_options,
                                             share_server=None):
        return self.manage_existing_snapshot(snapshot, driver_options)

    @slow_me_down
    def manage_existing_snapshot(self, snapshot, driver_options):
        """Brings an existing snapshot under Manila management."""
        old_snap_id = self._get_snap_id_from_provider_location(
            snapshot['provider_location'])
        old_provider_location = self.private_storage.get(
            old_snap_id, key='provider_location')
        if old_provider_location == snapshot['provider_location']:
            self._create_snapshot(snapshot)
            self.private_storage.delete(old_snap_id)
            return {"size": 1,
                    "provider_location": snapshot["provider_location"]}
        else:
            msg = ("Invalid provider location specified, existing snapshot %s"
                   " could not be found" % old_snap_id)
            raise exception.ShareBackendException(msg=msg)

    def _get_snap_id_from_provider_location(self, provider_location):
        values = provider_location.split('snapshot_')
        if len(values) > 1:
            return values[1][37:].replace("_", "-")
        else:
            return provider_location

    @slow_me_down
    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Manila management."""
        self.private_storage.update(
            snapshot['id'],
            {'provider_location': snapshot['provider_location']})

    @slow_me_down
    def unmanage_snapshot_with_server(self, snapshot, share_server=None):
        self.unmanage_snapshot(snapshot)

    @slow_me_down
    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Reverts a share (in place) to the specified snapshot."""

    @slow_me_down
    def extend_share(self, share, new_size, share_server=None):
        """Extends size of existing share."""

    @slow_me_down
    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks size of existing share."""

    def get_network_allocations_number(self):
        """Returns number of network allocations for creating VIFs."""
        return 2

    def get_admin_network_allocations_number(self):
        return 1

    @slow_me_down
    def _setup_server(self, network_info, metadata=None):
        """Sets up and configures share server with given network parameters.

        Redefine it within share driver when it is going to handle share
        servers.
        """
        common_net_info = network_info[0]
        server_details = {
            "service_ip": common_net_info[
                "admin_network_allocations"][0]["ip_address"],
            "username": "fake_username",
            "server_id": common_net_info['server_id'],
        }

        subnet_allocations = []
        for subnet_info in network_info:
            subnet_allocations.append({
                "primary_public_ip": subnet_info[
                    "network_allocations"][0]["ip_address"],
                "secondary_public_ip": subnet_info[
                    "network_allocations"][1]["ip_address"]
            })

        server_details['subnet_allocations'] = jsonutils.dumps(
            subnet_allocations)
        return server_details

    @slow_me_down
    def _teardown_server(self, server_details, security_services=None):
        """Tears down share server."""

    @slow_me_down
    def _get_pools_info(self):
        pools = [{
            "pool_name": "fake_pool_for_%s" % self.backend_name,
            "total_capacity_gb": 1230.0,
            "free_capacity_gb": 1210.0,
            "reserved_percentage":
                self.configuration.reserved_share_percentage,
            "reserved_snapshot_percentage":
                self.configuration.reserved_share_from_snapshot_percentage,
            "reserved_share_extend_percentage":
                self.configuration.reserved_share_extend_percentage
        }]
        if self.configuration.replication_domain:
            pools[0]["replication_type"] = "readable"
        return pools

    @slow_me_down
    def _update_share_stats(self, data=None):
        """Retrieve stats info from share group."""
        data = {
            "share_backend_name": self.backend_name,
            "storage_protocol": "NFS_CIFS",
            "reserved_percentage":
                self.configuration.reserved_share_percentage,
            "reserved_snapshot_percentage":
                self.configuration.reserved_share_from_snapshot_percentage,
            "reserved_share_extend_percentage":
                self.configuration.reserved_share_extend_percentage,
            "snapshot_support": True,
            "create_share_from_snapshot_support": True,
            "revert_to_snapshot_support": True,
            "mount_snapshot_support": True,
            "driver_name": "Dummy",
            "pools": self._get_pools_info(),
            "share_group_stats": {
                "consistent_snapshot_support": "pool",
            },
            'share_server_multiple_subnet_support': True,
        }
        if self.configuration.replication_domain:
            data["replication_type"] = "readable"
        super(DummyDriver, self)._update_share_stats(data)

    def get_share_server_pools(self, share_server):
        """Return list of pools related to a particular share server."""
        return []

    @slow_me_down
    def create_consistency_group(self, context, cg_dict, share_server=None):
        """Create a consistency group."""
        LOG.debug(
            "Successfully created dummy Consistency Group with ID: %s.",
            cg_dict["id"])

    @slow_me_down
    def delete_consistency_group(self, context, cg_dict, share_server=None):
        """Delete a consistency group."""
        LOG.debug(
            "Successfully deleted dummy consistency group with ID %s.",
            cg_dict["id"])

    @slow_me_down
    def create_cgsnapshot(self, context, snap_dict, share_server=None):
        """Create a consistency group snapshot."""
        LOG.debug("Successfully created CG snapshot %s.", snap_dict["id"])
        return None, None

    @slow_me_down
    def delete_cgsnapshot(self, context, snap_dict, share_server=None):
        """Delete a consistency group snapshot."""
        LOG.debug("Successfully deleted CG snapshot %s.", snap_dict["id"])
        return None, None

    @slow_me_down
    def create_consistency_group_from_cgsnapshot(
            self, context, cg_dict, cgsnapshot_dict, share_server=None):
        """Create a consistency group from a cgsnapshot."""
        LOG.debug(
            ("Successfully created dummy Consistency Group (%(cg_id)s) "
             "from CG snapshot (%(cg_snap_id)s)."),
            {"cg_id": cg_dict["id"], "cg_snap_id": cgsnapshot_dict["id"]})
        return None, []

    @slow_me_down
    def create_replica(self, context, replica_list, new_replica,
                       access_rules, replica_snapshots, share_server=None):
        """Replicate the active replica to a new replica on this backend."""
        replica_name = self._get_share_name(new_replica)
        mountpoint = "/path/to/fake/share/%s" % replica_name
        self.private_storage.update(
            new_replica["id"], {
                "fake_provider_replica_name": replica_name,
                "fake_provider_location": mountpoint,
            }
        )
        return {
            "export_locations": self._generate_export_locations(
                mountpoint, share_server=share_server),
            "replica_state": constants.REPLICA_STATE_IN_SYNC,
            "access_rules_status": constants.STATUS_ACTIVE,
        }

    @slow_me_down
    def delete_replica(self, context, replica_list, replica_snapshots,
                       replica, share_server=None):
        """Delete a replica."""
        self.private_storage.delete(replica["id"])

    @slow_me_down
    def promote_replica(self, context, replica_list, replica, access_rules,
                        share_server=None):
        """Promote a replica to 'active' replica state."""
        return_replica_list = []
        for r in replica_list:
            if r["id"] == replica["id"]:
                replica_state = constants.REPLICA_STATE_ACTIVE
            else:
                replica_state = constants.REPLICA_STATE_IN_SYNC
            return_replica_list.append(
                {"id": r["id"], "replica_state": replica_state})
        return return_replica_list

    @slow_me_down
    def update_replica_state(self, context, replica_list, replica,
                             access_rules, replica_snapshots,
                             share_server=None):
        """Update the replica_state of a replica."""
        return constants.REPLICA_STATE_IN_SYNC

    @slow_me_down
    def create_replicated_snapshot(self, context, replica_list,
                                   replica_snapshots, share_server=None):
        """Create a snapshot on active instance and update across the replicas.

        """
        return_replica_snapshots = []
        for r in replica_snapshots:
            return_replica_snapshots.append(
                {"id": r["id"], "status": constants.STATUS_AVAILABLE})
        return return_replica_snapshots

    @slow_me_down
    def revert_to_replicated_snapshot(self, context, active_replica,
                                      replica_list, active_replica_snapshot,
                                      replica_snapshots, share_access_rules,
                                      snapshot_access_rules,
                                      share_server=None):
        """Reverts a replicated share (in place) to the specified snapshot."""

    @slow_me_down
    def delete_replicated_snapshot(self, context, replica_list,
                                   replica_snapshots, share_server=None):
        """Delete a snapshot by deleting its instances across the replicas."""
        return_replica_snapshots = []
        for r in replica_snapshots:
            return_replica_snapshots.append(
                {"id": r["id"], "status": constants.STATUS_DELETED})
        return return_replica_snapshots

    @slow_me_down
    def update_replicated_snapshot(self, context, replica_list,
                                   share_replica, replica_snapshots,
                                   replica_snapshot, share_server=None):
        """Update the status of a snapshot instance that lives on a replica."""
        return {
            "id": replica_snapshot["id"], "status": constants.STATUS_AVAILABLE}

    @slow_me_down
    def migration_check_compatibility(
            self, context, source_share, destination_share,
            share_server=None, destination_share_server=None):
        """Is called to test compatibility with destination backend."""
        backend_name = share_utils.extract_host(
            destination_share['host'], level='backend_name')
        config = get_backend_configuration(backend_name)
        compatible = 'Dummy' in config.share_driver
        return {
            'compatible': compatible,
            'writable': compatible,
            'preserve_metadata': compatible,
            'nondisruptive': False,
            'preserve_snapshots': compatible,
        }

    @slow_me_down
    def migration_start(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to perform 1st phase of driver migration of a given share.

        """
        LOG.debug(
            "Migration of dummy share with ID '%s' has been started.",
            source_share["id"])
        self.migration_progress[source_share['share_id']] = 0

    @slow_me_down
    def migration_continue(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):

        if source_share["id"] not in self.migration_progress:
            self.migration_progress[source_share["id"]] = 0

        self.migration_progress[source_share["id"]] += 50

        LOG.debug(
            "Migration of dummy share with ID '%s' is continuing, %s.",
            source_share["id"],
            self.migration_progress[source_share["id"]])

        return self.migration_progress[source_share["id"]] == 100

    @slow_me_down
    def migration_complete(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to perform 2nd phase of driver migration of a given share.

        """
        snapshot_updates = {}
        for src_snap_ins, dest_snap_ins in snapshot_mappings.items():
            snapshot_updates[dest_snap_ins['id']] = self._create_snapshot(
                dest_snap_ins)
        return {
            'snapshot_updates': snapshot_updates,
            'export_locations': self._do_migration(
                source_share, destination_share, share_server)
        }

    def _do_migration(self, source_share_ref, dest_share_ref, share_server):
        share_name = self._get_share_name(dest_share_ref)
        mountpoint = "/path/to/fake/share/%s" % share_name
        self.private_storage.delete(source_share_ref["id"])
        self.private_storage.update(
            dest_share_ref["id"], {
                "fake_provider_share_name": share_name,
                "fake_provider_location": mountpoint,
            }
        )
        LOG.debug(
            "Migration of dummy share with ID '%s' has been completed.",
            source_share_ref["id"])
        self.migration_progress.pop(source_share_ref["id"], None)

        return self._generate_export_locations(
            mountpoint, share_server=share_server)

    @slow_me_down
    def migration_cancel(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to cancel driver migration."""
        LOG.debug(
            "Migration of dummy share with ID '%s' has been canceled.",
            source_share["id"])
        self.migration_progress.pop(source_share["id"], None)

    @slow_me_down
    def migration_get_progress(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to get migration progress."""
        # Simulate migration progress.
        if source_share["id"] not in self.migration_progress:
            self.migration_progress[source_share["id"]] = 0
        total_progress = self.migration_progress[source_share["id"]]
        LOG.debug("Progress of current dummy share migration "
                  "with ID '%(id)s' is %(progress)s.", {
                      "id": source_share["id"],
                      "progress": total_progress
                  })
        return {"total_progress": total_progress}

    def share_server_migration_check_compatibility(
            self, context, share_server, dest_host, old_share_network,
            new_share_network, shares_request_spec):
        """Is called to check migration compatibility for a share server."""
        backend_name = share_utils.extract_host(
            dest_host, level='backend_name')
        config = get_backend_configuration(backend_name)
        compatible = 'Dummy' in config.share_driver
        return {
            'compatible': compatible,
            'writable': compatible,
            'preserve_snapshots': compatible,
            'nondisruptive': False,
            'share_network_id': new_share_network['id'],
            'migration_cancel': compatible,
            'migration_get_progress': compatible,
        }

    @slow_me_down
    def share_server_migration_start(self, context, src_share_server,
                                     dest_share_server, shares, snapshots):
        """Is called to perform 1st phase of migration of a share server."""
        LOG.debug(
            "Migration of dummy share server with ID '%s' has been started.",
            src_share_server["id"])
        self.migration_progress[src_share_server['id']] = 0

    @slow_me_down
    def share_server_migration_continue(self, context, src_share_server,
                                        dest_share_server, shares, snapshots):
        """Is called to continue the migration of a share server."""
        if src_share_server["id"] not in self.migration_progress:
            self.migration_progress[src_share_server["id"]] = 0

        self.migration_progress[src_share_server["id"]] += 50

        LOG.debug(
            "Migration of dummy share server with ID '%s' is continuing, %s.",
            src_share_server["id"],
            self.migration_progress[src_share_server["id"]])

        return self.migration_progress[src_share_server["id"]] >= 100

    @slow_me_down
    def share_server_migration_complete(self, context, source_share_server,
                                        dest_share_server, shares, snapshots,
                                        new_network_allocations):
        """Is called to complete the migration of a share server."""
        shares_updates = {}
        pools = self._get_pools_info()
        for instance in shares:

            share_name = self._get_share_name(instance)
            mountpoint = "/path/to/fake/share/%s" % share_name
            export_locations = self._generate_export_locations(
                mountpoint, share_server=dest_share_server)
            dest_pool = pools[0]['pool_name']
            shares_updates.update(
                {instance['id']: {'export_locations': export_locations,
                                  'pool_name': dest_pool}}
            )

        snapshot_updates = {}
        for instance in snapshots:
            snapshot_name = self._get_snapshot_name(instance)
            mountpoint = "/path/to/fake/snapshot/%s" % snapshot_name
            snap_export_locations = self._generate_export_locations(
                mountpoint, share_server=dest_share_server)
            snapshot_updates.update(
                {instance['id']: {
                    'provider_location': mountpoint,
                    'export_locations': snap_export_locations}}
            )

        LOG.debug(
            "Migration of dummy share server with ID '%s' has been completed.",
            source_share_server["id"])
        self.migration_progress.pop(source_share_server["id"], None)

        return {
            'share_updates': shares_updates,
            'snapshot_updates': snapshot_updates,
        }

    @slow_me_down
    def share_server_migration_cancel(self, context, src_share_server,
                                      dest_share_server, shares, snapshots):
        """Is called to cancel a share server migration."""
        LOG.debug(
            "Migration of dummy share server with ID '%s' has been canceled.",
            src_share_server["id"])
        self.migration_progress.pop(src_share_server["id"], None)

    @slow_me_down
    def share_server_migration_get_progress(self, context, src_share_server,
                                            dest_share_server, shares,
                                            snapshots):
        """Is called to get share server migration progress."""
        if src_share_server["id"] not in self.migration_progress:
            self.migration_progress[src_share_server["id"]] = 0
        total_progress = self.migration_progress[src_share_server["id"]]
        LOG.debug("Progress of current dummy share server migration "
                  "with ID '%(id)s' is %(progress)s.", {
                      "id": src_share_server["id"],
                      "progress": total_progress
                  })
        return {"total_progress": total_progress}

    def update_share_usage_size(self, context, shares):
        share_updates = []
        gathered_at = timeutils.utcnow()
        for s in shares:
            share_updates.append({'id': s['id'],
                                  'used_size': 1,
                                  'gathered_at': gathered_at})
        return share_updates

    @slow_me_down
    def get_share_server_network_info(
            self, context, share_server, identifier, driver_options):
        try:
            server_details = self.private_storage.get(identifier)
        except Exception:
            msg = ("Unable to find share server %s in "
                   "private storage." % identifier)
            raise exception.ShareBackendException(msg=msg)

        ips = [server_details['service_ip']]

        subnet_allocations = (
            self._get_subnet_allocations_from_backend_details(server_details))

        for subnet_allocation in subnet_allocations:
            ips += list(subnet_allocation.values())
        return ips

    @slow_me_down
    def manage_server(self, context, share_server, identifier, driver_options):
        server_details = self.private_storage.get(identifier)
        self.private_storage.delete(identifier)
        return identifier, server_details

    def unmanage_server(self, server_details, security_services=None):
        server_details = server_details or {}
        if not server_details or 'server_id' not in server_details:
            # This share server doesn't have any network details. Since it's
            # just being cleaned up, we'll log a warning and return without
            # errors.
            LOG.warning("Share server does not have network information. "
                        "It is being unmanaged, but cannot be re-managed "
                        "without first creating network allocations in this "
                        "driver's private storage.")
            return
        self.private_storage.update(server_details['server_id'],
                                    server_details)

    def get_share_status(self, share, share_server=None):
        return {
            'status': constants.STATUS_AVAILABLE,
            'export_locations': self.private_storage.get(share['id'],
                                                         key='export_location')
        }

    @slow_me_down
    def update_share_server_security_service(self, context, share_server,
                                             network_info, share_instances,
                                             share_instance_rules,
                                             new_security_service,
                                             current_security_service=None):
        if current_security_service:
            msg = _("Replacing security service %(cur_sec_serv_id)s by "
                    "security service %(new_sec_serv_id)s on share server "
                    "%(server_id)s."
                    ) % {
                'cur_sec_serv_id': current_security_service['id'],
                'new_sec_serv_id': new_security_service['id'],
                'server_id': share_server['id']
            }
        else:
            msg = _("Adding security service %(sec_serv_id)s on share server "
                    "%(server_id)s."
                    ) % {
                'sec_serv_id': new_security_service['id'],
                'server_id': share_server['id']
            }

        LOG.debug(msg)

    def check_update_share_server_security_service(
            self, context, share_server, network_info, share_instances,
            share_instance_rules, new_security_service,
            current_security_service=None):
        return True

    def check_update_share_server_network_allocations(
            self, context, share_server, current_network_allocations,
            new_share_network_subnet, security_services, share_instances,
            share_instances_rules):

        LOG.debug("Share server %(server)s can be updated with allocations "
                  "from new subnet.", {'server': share_server['id']})
        return True

    def update_share_server_network_allocations(
            self, context, share_server, current_network_allocations,
            new_network_allocations, security_services, shares, snapshots):

        backend_details = share_server['backend_details']
        subnet_allocations = (
            self._get_subnet_allocations_from_backend_details(backend_details))

        subnet_allocations.append({
            'primary_public_ip': new_network_allocations[
                'network_allocations'][0]['ip_address'],
            'secondary_public_ip': new_network_allocations[
                'network_allocations'][1]['ip_address'],
        })
        new_server = {
            "backend_details": {
                "subnet_allocations": jsonutils.dumps(subnet_allocations),
                "service_ip": backend_details["service_ip"],
            }
        }
        shares_updates = {}
        for instance in shares:

            share_name = self._get_share_name(instance)
            mountpoint = "/path/to/fake/share/%s" % share_name
            export_locations = self._generate_export_locations(
                mountpoint, share_server=new_server)
            shares_updates.update(
                {instance['id']: export_locations}
            )

        snapshot_updates = {}
        for instance in snapshots:
            snapshot_name = self._get_snapshot_name(instance)
            mountpoint = "/path/to/fake/snapshot/%s" % snapshot_name
            snap_export_locations = self._generate_export_locations(
                mountpoint, share_server=new_server)
            snapshot_updates.update(
                {instance['id']: {
                    'provider_location': mountpoint,
                    'export_locations': snap_export_locations}}
            )

        LOG.debug(
            "Network update allocations of dummy share server with ID '%s' "
            "has been completed.", share_server["id"])
        return {
            "share_updates": shares_updates,
            "snapshot_updates": snapshot_updates,
            "server_details": {
                "subnet_allocations": (
                    new_server["backend_details"]["subnet_allocations"])
            },
        }



# Copyright 2021 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila import exception
from manila import utils as manila_utils
from oslo_log import log

LOG = log.getLogger(__name__)


def execute_with_retries(action, action_args, max_retries):

    @manila_utils.retry(
        retry_param=exception.ProcessExecutionError, backoff_rate=2,
        retries=max_retries)
    def execute():
        try:
            action(*action_args)
            return True
        except exception.ProcessExecutionError:
            LOG.exception("Recovering from a failed execute.")
            raise

    try:
        execute()
    except exception.ProcessExecutionError:
        LOG.exception("Failed to run command. Tries exhausted.")
        raise



# Copyright 2021 Cloudification GmbH.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.scheduler.filters import base_host


class OnlyHostFilter(base_host.BaseHostFilter):
    """Filters Hosts by 'only_host' scheduler_hint."""

    def host_passes(self, host_state, filter_properties):
        context = filter_properties['context']
        if not context.is_admin:
            return True
        hints = filter_properties.get('scheduler_hints')
        if hints is None:
            return True
        requested_host = hints.get('only_host', None)
        if requested_host is None:
            return True
        # e.g. "only_host=hostname@generic2#GENERIC2"
        return host_state.host == requested_host



# Copyright (c) 2016 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unit tests for the Container helper module."""

from unittest import mock

import ddt

from manila import exception
from manila.share import configuration
from manila.share.drivers.container import container_helper
from manila import test
from manila.tests.share.drivers.container import fakes


@ddt.ddt
class DockerExecHelperTestCase(test.TestCase):
    """Tests DockerExecHelper"""

    def setUp(self):
        super(DockerExecHelperTestCase, self).setUp()
        self.fake_conf = configuration.Configuration(None)
        self.fake_conf.container_image_name = "fake_image"
        self.fake_conf.container_volume_mount_path = "/tmp/shares"
        self.DockerExecHelper = container_helper.DockerExecHelper(
            configuration=self.fake_conf)

    def test_create_container(self):
        fake_name = 'fake_container'
        self.DockerExecHelper.configuration.container_image_name = 'fake_image'

        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(return_value=('fake_container_id', '')))
        self.mock_object(self.DockerExecHelper, 'disconnect_network')

        self.DockerExecHelper.create_container(fake_name)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'container', 'create', '--name=%s' % fake_name,
            '--privileged', '-v', '/dev:/dev', '-v', '/tmp/shares:/shares',
            'fake_image'])
        self.DockerExecHelper.disconnect_network.assert_called_once_with(
            'bridge', fake_name)

    def test_create_container_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.create_container)

    def test_start_container(self):
        fake_name = 'fake_container'

        self.mock_object(self.DockerExecHelper, '_inner_execute', mock.Mock())

        self.DockerExecHelper.start_container(fake_name)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'container', 'start', 'fake_container'])

    def test_start_container_impossible_failure(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.start_container, None)

    def test_stop_container(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(return_value=['fake_output', None]))
        expected = ['docker', 'stop', 'manila-fake-conainer']

        self.DockerExecHelper.stop_container("manila-fake-conainer")

        self.DockerExecHelper._inner_execute.assert_called_once_with(expected)

    def test_stop_container_oh_noes(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(side_effect=OSError))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.stop_container,
                          "manila-fake-container")

    def test_execute(self):
        self.mock_object(self.DockerExecHelper, "_inner_execute",
                         mock.Mock(return_value='fake_output'))
        expected = ['docker', 'exec', '-i', 'fake_container', 'fake_script']

        self.DockerExecHelper.execute("fake_container", ["fake_script"])

        self.DockerExecHelper._inner_execute.assert_called_once_with(
            expected, ignore_errors=False)

    def test_execute_name_not_there(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          None, ['do', 'stuff'])

    def test_execute_command_not_there(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          'fake-name', None)

    def test_execute_bad_command_format(self):
        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.execute,
                          'fake-name', 'do stuff')

    def test__inner_execute_ok(self):
        self.DockerExecHelper._execute = mock.Mock(return_value='fake')

        result = self.DockerExecHelper._inner_execute("fake_command")

        self.assertEqual(result, 'fake')

    def test__inner_execute_not_ok(self):
        self.DockerExecHelper._execute = mock.Mock(side_effect=[OSError()])

        self.assertRaises(OSError,
                          self.DockerExecHelper._inner_execute, "fake_command")

    def test__inner_execute_not_ok_ignore_errors(self):
        self.DockerExecHelper._execute = mock.Mock(side_effect=OSError())

        result = self.DockerExecHelper._inner_execute("fake_command",
                                                      ignore_errors=True)

        self.assertIsNone(result)

    def test_fetch_container_addresses(self):
        fake_name = 'fake_container'
        fake_addresses = ['192.168.144.19', '10.0.0.131']
        fake_ip_addr_show = fakes.FAKE_IP_ADDR_SHOW
        fake_interfaces = ['eth0', 'eth1']

        self.mock_object(self.DockerExecHelper, 'fetch_container_interfaces',
                         mock.Mock(return_value=fake_interfaces))
        self.mock_object(self.DockerExecHelper, 'execute',
                         mock.Mock(side_effect=[fake_ip_addr_show[0],
                                                fake_ip_addr_show[1]]))

        self.assertEqual(fake_addresses,
                         self.DockerExecHelper.fetch_container_addresses(
                             fake_name, 'inet'))
        (self.DockerExecHelper.fetch_container_interfaces
         .assert_called_once_with(fake_name))
        self.DockerExecHelper.execute.assert_any_call(
            fake_name, ['ip', '-oneline', '-family', 'inet', 'address',
                        'show', 'scope', 'global', 'dev', 'eth0']
        )
        self.DockerExecHelper.execute.assert_any_call(
            fake_name, ['ip', '-oneline', '-family', 'inet', 'address',
                        'show', 'scope', 'global', 'dev', 'eth1']
        )

    def test_fetch_container_interfaces(self):
        fake_name = 'fake_container'
        fake_eths = fakes.FAKE_IP_LINK_SHOW

        self.mock_object(self.DockerExecHelper, 'execute',
                         mock.Mock(return_value=fake_eths))

        self.assertEqual(
            ['eth0', 'eth1'],
            self.DockerExecHelper.fetch_container_interfaces(fake_name))
        self.DockerExecHelper.execute.assert_called_once_with(
            fake_name, ['ip', '-o', 'link', 'show'])

    def test_rename_container(self):
        fake_old_name = 'old_name'
        fake_new_name = 'new_name'
        fake_veth_names = ['fake_veth']

        self.mock_object(self.DockerExecHelper, 'get_container_veths',
                         mock.Mock(return_value=fake_veth_names))
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=[None, None]))

        self.DockerExecHelper.rename_container(fake_old_name, fake_new_name)

        self.DockerExecHelper.get_container_veths.assert_called_once_with(
            fake_old_name)
        self.DockerExecHelper._inner_execute.assert_has_calls([
            mock.call(['docker', 'rename', fake_old_name, fake_new_name]),
            mock.call(['ovs-vsctl', 'set', 'interface', fake_veth_names[0],
                       'external-ids:manila-container=%s' % fake_new_name])])

    def test_rename_container_exception_veth(self):
        fake_old_name = 'old_name'
        fake_new_name = 'new_name'

        self.mock_object(self.DockerExecHelper, 'get_container_veths',
                         mock.Mock(return_value=[]))

        self.assertRaises(exception.ManilaException,
                          self.DockerExecHelper.rename_container,
                          fake_old_name, fake_new_name)

    @ddt.data([['fake', ''], OSError, ['fake', '']],
              [['fake', ''], OSError, OSError],
              [OSError])
    def test_rename_container_exception_cmds(self, side_effect):
        fake_old_name = 'old_name'
        fake_new_name = 'new_name'
        fake_veth_names = ['fake_veth']

        self.mock_object(self.DockerExecHelper, 'get_container_veths',
                         mock.Mock(return_value=fake_veth_names))
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=side_effect))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.rename_container,
                          fake_old_name, fake_new_name)

        if len(side_effect) > 1:
            self.DockerExecHelper._inner_execute.assert_has_calls([
                mock.call(['docker', 'rename', fake_old_name, fake_new_name]),
                mock.call(['ovs-vsctl', 'set', 'interface', fake_veth_names[0],
                           'external-ids:manila-container=%s' % fake_new_name])
            ])
        else:
            self.DockerExecHelper._inner_execute.assert_has_calls([
                mock.call(['docker', 'rename', fake_old_name, fake_new_name])])

    @ddt.data((["wrong_name\nfake\nfake_container\nfake_name'"], True),
              (["wrong_name\nfake_container\nfake'"], False),
              ("\n", False))
    @ddt.unpack
    def test_container_exists(self, fake_return_value, expected_result):

        self.DockerExecHelper._execute = mock.Mock(
            return_value=fake_return_value)

        result = self.DockerExecHelper.container_exists("fake_name")

        self.DockerExecHelper._execute.assert_called_once_with(
            "docker", "ps", "--no-trunc", "--format='{{.Names}}'",
            run_as_root=True)
        self.assertEqual(expected_result, result)

    def test_create_network(self):
        fake_network_name = 'fake_network_name'

        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(return_value=('fake_network_id', '')))

        self.DockerExecHelper.create_network(fake_network_name)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'network', 'create', fake_network_name])

    def test_create_network_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.create_network, None)

    def test_remove_network(self):
        fake_network_name = 'fake_network_name'

        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(return_value=('fake_network_id', '')))

        self.DockerExecHelper.remove_network(fake_network_name)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'network', 'remove', fake_network_name])

    def test_remove_network_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.remove_network, None)

    def test_connect_network(self):
        fake_network_name = 'fake_network_name'
        fake_server_id = 'fake_server_id'

        self.mock_object(self.DockerExecHelper, '_inner_execute')

        self.DockerExecHelper.connect_network(fake_network_name,
                                              fake_server_id)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'network', 'connect', fake_network_name, fake_server_id])

    def test_connect_network_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.connect_network, None, None)

    def test_disconnect_network(self):
        fake_network_name = 'fake_network_name'
        fake_server_id = 'fake_server_id'

        self.mock_object(self.DockerExecHelper, '_inner_execute')

        self.DockerExecHelper.disconnect_network(fake_network_name,
                                                 fake_server_id)

        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'network', 'disconnect', fake_network_name,
            fake_server_id])

    def test_disconnect_network_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.disconnect_network, None, None)

    def test_get_container_networks(self):
        fake_container_name = 'fake_container_name'
        fake_docker_inspect_networks = fakes.FAKE_DOCKER_INSPECT_NETWORKS
        fake_networks = ['fake_docker_network_0', 'fake_docker_network_1']

        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(return_value=fake_docker_inspect_networks))

        self.assertEqual(
            fake_networks,
            self.DockerExecHelper.get_container_networks(fake_container_name))
        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'container', 'inspect', '-f',
            '\'{{json .NetworkSettings.Networks}}\'', fake_container_name])

    def test_get_container_networks_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.get_container_networks, None)

    def test_get_container_veths(self):
        fake_container_name = 'fake_container_name'
        fake_eths_iflinks = ('10\n11\n', '')
        fake_veths = ['fake_veth_0', 'fake_veth_1']

        self.mock_object(self.DockerExecHelper, 'execute',
                         mock.Mock(return_value=fake_eths_iflinks))
        self.mock_object(
            self.DockerExecHelper, '_execute',
            mock.Mock(side_effect=[('/sys/class/net/%s/ifindex'
                                    % fake_veths[0], ''),
                                   ('/sys/class/net/%s/ifindex'
                                    % fake_veths[1], '')]))

        self.assertEqual(
            fake_veths,
            self.DockerExecHelper.get_container_veths(fake_container_name))
        self.DockerExecHelper.execute.assert_called_once_with(
            fake_container_name,
            ['bash', '-c', 'cat /sys/class/net/eth*/iflink'])
        self.DockerExecHelper._execute.assert_has_calls([
            mock.call('bash', '-c', 'grep -l 10 /sys/class/net/veth*/ifindex'),
            mock.call('bash', '-c', 'grep -l 11 /sys/class/net/veth*/ifindex')
        ])

    def test_get_network_bridge(self):
        fake_network_name = 'fake_network_name'
        fake_network_id = ('012345abcdef', '')
        fake_bridge = 'br-' + fake_network_id[0]

        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(return_value=fake_network_id))

        self.assertEqual(
            fake_bridge,
            self.DockerExecHelper.get_network_bridge(fake_network_name))
        self.DockerExecHelper._inner_execute.assert_called_once_with([
            'docker', 'network', 'inspect', '-f', '{{.Id}}', fake_network_name
        ])

    def test_get_network_bridge_failure(self):
        self.mock_object(self.DockerExecHelper, '_inner_execute',
                         mock.Mock(side_effect=OSError()))

        self.assertRaises(exception.ShareBackendException,
                          self.DockerExecHelper.get_network_bridge, None)

    def test_get_veth_from_bridge(self):
        fake_bridge = 'br-012345abcdef'
        fake_ip_link_show_master = fakes.FAKE_IP_LINK_SHOW_MASTER
        fake_veth = 'fake_veth'

        self.mock_object(self.DockerExecHelper, '_execute',
                         mock.Mock(return_value=fake_ip_link_show_master))

        self.assertEqual(
            fake_veth, self.DockerExecHelper.get_veth_from_bridge(fake_bridge))
        self.DockerExecHelper._execute.assert_called_once_with('ip', 'link',
                                                               'show',
                                                               'master',
                                                               fake_bridge)



# Copyright 2021 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Helpers for os basic commands
"""

from oslo_concurrency import processutils

from manila import exception

import manila.privsep


@manila.privsep.sys_admin_pctxt.entrypoint
def rmdir(dir_path):
    processutils.execute('rmdir', dir_path)


@manila.privsep.sys_admin_pctxt.entrypoint
def is_data_definition_direct_io_supported(src_str, dest_str):
    try:
        processutils.execute(
            'dd', 'count=0', f'if={src_str}', f'of={dest_str}',
            'iflag=direct', 'oflag=direct')
        is_direct_io_supported = True
    except exception.ProcessExecutionError:
        is_direct_io_supported = False

    return is_direct_io_supported


@manila.privsep.sys_admin_pctxt.entrypoint
def data_definition(src_str, dest_str, size_in_g, use_direct_io=False):
    extra_flags = []
    if use_direct_io:
        extra_flags += ['iflag=direct', 'oflag=direct']
    processutils.execute(
        'dd', 'if=%s' % src_str, 'of=%s' % dest_str, 'count=%d' % size_in_g,
        'bs=1M', *extra_flags)


@manila.privsep.sys_admin_pctxt.entrypoint
def umount(mount_path):
    processutils.execute('umount', '-f', mount_path)


@manila.privsep.sys_admin_pctxt.entrypoint
def mount(device_name, mount_path):
    processutils.execute('mount', device_name, mount_path)


@manila.privsep.sys_admin_pctxt.entrypoint
def list_mounts():
    out, err = processutils.execute('mount', '-l')
    return out, err


@manila.privsep.sys_admin_pctxt.entrypoint
def chmod(permission_level_str, mount_path):
    processutils.execute('chmod', permission_level_str, mount_path)



# Copyright 2019 NetApp, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client

from manila.api import common
from oslo_db import exception as db_exception
from oslo_log import log
import webob
from webob import exc

from manila.api.openstack import api_version_request as api_version
from manila.api.openstack import wsgi
from manila.api.views import share_network_subnets as subnet_views
from manila.db import api as db_api
from manila import exception
from manila.i18n import _
from manila import share
from manila.share import rpcapi as share_rpcapi

LOG = log.getLogger(__name__)


class ShareNetworkSubnetController(wsgi.Controller):
    """The Share Network Subnet API controller for the OpenStack API."""

    resource_name = 'share_network_subnet'
    _view_builder_class = subnet_views.ViewBuilder

    def __init__(self):
        super(ShareNetworkSubnetController, self).__init__()
        self.share_rpcapi = share_rpcapi.ShareAPI()
        self.share_api = share.API()

    @wsgi.Controller.api_version("2.51")
    @wsgi.Controller.authorize
    def index(self, req, share_network_id):
        """Returns a list of share network subnets."""
        context = req.environ['manila.context']

        try:
            share_network = db_api.share_network_get(context, share_network_id)
        except exception.ShareNetworkNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        return self._view_builder.build_share_network_subnets(
            req, share_network.get('share_network_subnets'))

    def _all_share_servers_are_auto_deletable(self, share_network_subnet):
        return all([ss['is_auto_deletable'] for ss
                    in share_network_subnet['share_servers']])

    @wsgi.Controller.api_version('2.51')
    @wsgi.Controller.authorize
    def delete(self, req, share_network_id, share_network_subnet_id):
        """Delete specified share network subnet."""
        context = req.environ['manila.context']

        try:
            db_api.share_network_get(context, share_network_id)
        except exception.ShareNetworkNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        try:
            share_network_subnet = db_api.share_network_subnet_get(
                context, share_network_subnet_id)
        except exception.ShareNetworkSubnetNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        for share_server in share_network_subnet['share_servers'] or []:
            shares = db_api.share_instances_get_all_by_share_server(
                context, share_server['id'])
            if shares:
                msg = _("Cannot delete share network subnet %(id)s, it has "
                        "one or more shares.") % {
                    'id': share_network_subnet_id}
                LOG.error(msg)
                raise exc.HTTPConflict(explanation=msg)

        # NOTE(silvacarlose): Do not allow the deletion of any share server
        # if any of them has the flag is_auto_deletable = False
        if not self._all_share_servers_are_auto_deletable(
                share_network_subnet):
            msg = _("The service cannot determine if there are any "
                    "non-managed shares on the share network subnet %(id)s,"
                    "so it cannot be deleted. Please contact the cloud "
                    "administrator to rectify.") % {
                'id': share_network_subnet_id}
            LOG.error(msg)
            raise exc.HTTPConflict(explanation=msg)

        for share_server in share_network_subnet['share_servers']:
            self.share_rpcapi.delete_share_server(context, share_server)

        db_api.share_network_subnet_delete(context, share_network_subnet_id)
        return webob.Response(status_int=http_client.ACCEPTED)

    @wsgi.Controller.api_version("2.51")
    @wsgi.Controller.authorize
    def create(self, req, share_network_id, body):
        """Add a new share network subnet into the share network."""
        context = req.environ['manila.context']
        if not self.is_valid_body(body, 'share-network-subnet'):
            msg = _("Share Network Subnet is missing from the request body.")
            raise exc.HTTPBadRequest(explanation=msg)
        data = body['share-network-subnet']
        data['share_network_id'] = share_network_id
        multiple_subnet_support = (req.api_version_request >=
                                   api_version.APIVersionRequest("2.70"))
        share_network, existing_subnets = common.validate_subnet_create(
            context, share_network_id, data, multiple_subnet_support)

        # create subnet operation on subnets with share servers means that an
        # allocation update is requested.
        if existing_subnets and existing_subnets[0]['share_servers']:

            # NOTE(felipe_rodrigues): all subnets have the same set of share
            # servers, so we can just get the servers from one of them. Not
            # necessarily all share servers from the specified AZ will be
            # updated, only the ones created with subnets in the AZ. Others
            # created with default AZ will only have its allocations updated
            # when default subnet set is updated.
            data['share_servers'] = existing_subnets[0]['share_servers']
            try:
                share_network_subnet = (
                    self.share_api.update_share_server_network_allocations(
                        context, share_network, data))
            except exception.ServiceIsDown as e:
                msg = _('Could not add the share network subnet.')
                LOG.error(e)
                raise exc.HTTPInternalServerError(explanation=msg)
            except exception.InvalidShareNetwork as e:
                raise exc.HTTPBadRequest(explanation=e.msg)
            except db_exception.DBError as e:
                msg = _('Could not add the share network subnet.')
                LOG.error(e)
                raise exc.HTTPInternalServerError(explanation=msg)
        else:
            try:
                share_network_subnet = db_api.share_network_subnet_create(
                    context, data)
            except db_exception.DBError as e:
                msg = _('Could not create the share network subnet.')
                LOG.error(e)
                raise exc.HTTPInternalServerError(explanation=msg)

        share_network_subnet = db_api.share_network_subnet_get(
            context, share_network_subnet['id'])
        return self._view_builder.build_share_network_subnet(
            req, share_network_subnet)

    @wsgi.Controller.api_version('2.51')
    @wsgi.Controller.authorize
    def show(self, req, share_network_id, share_network_subnet_id):
        """Show share network subnet."""
        context = req.environ['manila.context']

        try:
            db_api.share_network_get(context, share_network_id)
        except exception.ShareNetworkNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        try:
            share_network_subnet = db_api.share_network_subnet_get(
                context, share_network_subnet_id)
        except exception.ShareNetworkSubnetNotFound as e:
            raise exc.HTTPNotFound(explanation=e.msg)

        return self._view_builder.build_share_network_subnet(
            req, share_network_subnet)


def create_resource():
    return wsgi.Resource(ShareNetworkSubnetController())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import ddt
import webob

from manila.api.v2 import metadata
from manila import context
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


@ddt.ddt
class MetadataAPITest(test.TestCase):

    def _get_request(self, version="2.65", use_admin_context=True):
        req = fakes.HTTPRequest.blank(
            '/v2/shares/{resource_id}/metadata',
            version=version, use_admin_context=use_admin_context)
        return req

    def setUp(self):
        super(MetadataAPITest, self).setUp()
        self.controller = (
            metadata.MetadataController())
        self.controller.resource_name = 'share'
        self.admin_context = context.RequestContext('admin', 'fake', True)
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.resource = db_utils.create_share(size=1)

    def test_create_index_metadata(self):
        url = self._get_request()
        body = {'metadata': {'test_key1': 'test_v1', 'test_key2': 'test_v2'}}
        update = self.controller._create_metadata(
            url, self.resource['id'], body=body)

        get = self.controller._index_metadata(url, self.resource['id'])

        self.assertEqual(2, len(get['metadata']))
        self.assertEqual(update['metadata'], get['metadata'])

    @ddt.data(({'metadata': {'key1': 'v1'}}, 'key1'),
              ({'metadata': {'test_key1': 'test_v1'}}, 'test_key1'),
              ({'metadata': {'key1': 'v2'}}, 'key1'))
    @ddt.unpack
    def test_update_get_metadata_item(self, body, key):
        url = self._get_request()
        update = self.controller._update_metadata_item(
            url, self.resource['id'], body=body, key=key)
        self.assertEqual(body, update)

        get = self.controller._index_metadata(url, self.resource['id'])

        self.assertEqual(1, len(get))
        self.assertEqual(body['metadata'], get['metadata'])

        get_item = self.controller._show_metadata(url, self.resource['id'],
                                                  key=key)
        self.assertEqual({'meta': body['metadata']}, get_item)

    @ddt.data({'metadata': {'key1': 'v1', 'key2': 'v2'}},
              {'metadata': {'test_key1': 'test_v1'}},
              {'metadata': {'key1': 'v2'}})
    def test_update_all_metadata(self, body):
        url = self._get_request()
        update = self.controller._update_all_metadata(
            url, self.resource['id'], body=body)
        self.assertEqual(body, update)

        get = self.controller._index_metadata(url, self.resource['id'])

        self.assertEqual(len(body['metadata']), len(get['metadata']))
        self.assertEqual(body['metadata'], get['metadata'])

    def test_delete_metadata(self):
        body = {'metadata': {'test_key3': 'test_v3', 'testkey': 'testval'}}
        url = self._get_request()
        self.controller._create_metadata(url, self.resource['id'], body=body)

        self.controller._delete_metadata(url, self.resource['id'],
                                         'test_key3')
        show_result = self.controller._index_metadata(url, self.resource[
            'id'])

        self.assertEqual(1, len(show_result['metadata']))
        self.assertNotIn('test_key3', show_result['metadata'])

    def test_update_metadata_with_resource_id_not_found(self):
        url = self._get_request()
        id = 'invalid_id'
        body = {'metadata': {'key1': 'v1'}}

        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller._create_metadata,
            url, id, body)

    def test_update_metadata_with_body_error(self):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._create_metadata,
            self._get_request(), self.resource['id'],
            {'metadata_error': {'key1': 'v1'}})

    @ddt.data({'metadata': {'key1': 'v1', 'key2': None}},
              {'metadata': {None: 'v1', 'key2': 'v2'}},
              {'metadata': {'k' * 256: 'v2'}},
              {'metadata': {'key1': 'v' * 1024}})
    @ddt.unpack
    def test_update_metadata_with_invalid_metadata(self, metadata):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._create_metadata,
            self._get_request(), self.resource['id'],
            {'metadata': metadata})

    def test_delete_metadata_not_found(self):
        body = {'metadata': {'test_key_exist': 'test_v_exist'}}
        update = self.controller._update_all_metadata(
            self._get_request(), self.resource['id'], body=body)
        self.assertEqual(body, update)
        self.assertRaises(
            exception.MetadataItemNotFound,
            self.controller._delete_metadata,
            self._get_request(), self.resource['id'], 'key1')



# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
from oslo_log import log
import webob
from webob import exc

from manila.api import common as api_common
from manila.api.openstack import wsgi
from manila.common import constants
from manila import db
from manila import exception
from manila.i18n import _
from manila import policy
from manila import share


LOG = log.getLogger(__name__)


class ShareMetadataController(object):
    """The share metadata API controller for the OpenStack API."""

    def __init__(self):
        self.share_api = share.API()
        super(ShareMetadataController, self).__init__()

    def _get_metadata(self, context, share_id):
        try:
            share = self.share_api.get(context, share_id)
            rv = db.share_metadata_get(context, share['id'])
            meta = dict(rv.items())
        except exception.NotFound:
            msg = _('share does not exist')
            raise exc.HTTPNotFound(explanation=msg)
        return meta

    def index(self, req, share_id):
        """Returns the list of metadata for a given share."""
        context = req.environ['manila.context']
        return {'metadata': self._get_metadata(context, share_id)}

    def create(self, req, share_id, body):
        try:
            metadata = body['metadata']
        except (KeyError, TypeError):
            msg = _("Malformed request body")
            raise exc.HTTPBadRequest(explanation=msg)

        context = req.environ['manila.context']

        new_metadata = self._update_share_metadata(context,
                                                   share_id,
                                                   metadata,
                                                   delete=False)
        return {'metadata': new_metadata}

    def update(self, req, share_id, id, body):
        try:
            meta_item = body['meta']
        except (TypeError, KeyError):
            expl = _('Malformed request body')
            raise exc.HTTPBadRequest(explanation=expl)

        if id not in meta_item:
            expl = _('Request body and URI mismatch')
            raise exc.HTTPBadRequest(explanation=expl)

        if len(meta_item) > 1:
            expl = _('Request body contains too many items')
            raise exc.HTTPBadRequest(explanation=expl)

        context = req.environ['manila.context']
        self._update_share_metadata(context,
                                    share_id,
                                    meta_item,
                                    delete=False)

        return {'meta': meta_item}

    def update_all(self, req, share_id, body):
        try:
            metadata = body['metadata']
        except (TypeError, KeyError):
            expl = _('Malformed request body')
            raise exc.HTTPBadRequest(explanation=expl)

        context = req.environ['manila.context']

        new_metadata = self._update_share_metadata(
            context, share_id, metadata, delete=True)
        return {'metadata': new_metadata}

    def _update_share_metadata(self, context,
                               share_id, metadata,
                               delete=False):
        ignore_keys = constants.AdminOnlyMetadata.SCHEDULER_FILTERS
        try:
            share = self.share_api.get(context, share_id)
            if set(metadata).intersection(set(ignore_keys)):
                try:
                    policy.check_policy(
                        context, 'share', 'update_admin_only_metadata')
                except exception.PolicyNotAuthorized:
                    msg = _("Cannot set or update admin only metadata.")
                    LOG.exception(msg)
                    raise exc.HTTPForbidden(explanation=msg)
                ignore_keys = []

            rv = db.share_metadata_get(context, share['id'])
            orig_meta = dict(rv.items())
            if delete:
                _metadata = metadata
                for key in ignore_keys:
                    if key in orig_meta:
                        _metadata[key] = orig_meta[key]
            else:
                metadata_copy = metadata.copy()
                for key in ignore_keys:
                    metadata_copy.pop(key, None)
                _metadata = orig_meta.copy()
                _metadata.update(metadata_copy)

            api_common._check_metadata_properties(_metadata)
            db.share_metadata_update(context, share['id'],
                                     _metadata, delete)

            return _metadata
        except exception.NotFound:
            msg = _('share does not exist')
            raise exc.HTTPNotFound(explanation=msg)

        except (ValueError, AttributeError):
            msg = _("Malformed request body")
            raise exc.HTTPBadRequest(explanation=msg)

        except exception.InvalidMetadata as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

        except exception.InvalidMetadataSize as error:
            raise exc.HTTPBadRequest(explanation=error.msg)

    def show(self, req, share_id, id):
        """Return a single metadata item."""
        context = req.environ['manila.context']
        data = self._get_metadata(context, share_id)

        try:
            return {'meta': {id: data[id]}}
        except KeyError:
            msg = _("Metadata item was not found")
            raise exc.HTTPNotFound(explanation=msg)

    def delete(self, req, share_id, id):
        """Deletes an existing metadata."""
        context = req.environ['manila.context']

        metadata = self._get_metadata(context, share_id)

        if id not in metadata:
            msg = _("Metadata item was not found")
            raise exc.HTTPNotFound(explanation=msg)

        try:
            share = self.share_api.get(context, share_id)
            if id in constants.AdminOnlyMetadata.SCHEDULER_FILTERS:
                policy.check_policy(context, 'share',
                                    'update_admin_only_metadata')
            db.share_metadata_delete(context, share['id'], id)
        except exception.NotFound:
            msg = _('share does not exist')
            raise exc.HTTPNotFound(explanation=msg)
        except exception.PolicyNotAuthorized:
            msg = _("Cannot delete admin only metadata.")
            LOG.exception(msg)
            raise exc.HTTPForbidden(explanation=msg)
        return webob.Response(status_int=http_client.OK)


def create_resource():
    return wsgi.Resource(ShareMetadataController())



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck

from manila.i18n import _


class Checks(upgradecheck.UpgradeCommands):

    """Various upgrade checks should be added as separate methods in this class

    and added to _upgrade_checks tuple.
    """

    def _check_placeholder(self):
        # This is just a placeholder for upgrade checks, it should be
        # removed when the actual checks are added
        return upgradecheck.Result(upgradecheck.Code.SUCCESS)

    # The format of the check functions is to return an
    # oslo_upgradecheck.upgradecheck.Result
    # object with the appropriate
    # oslo_upgradecheck.upgradecheck.Code and details set.
    # If the check hits warnings or failures then those should be stored
    # in the returned Result's "details" attribute. The
    # summary will be rolled up at the end of the check() method.
    _upgrade_checks = (
        # In the future there should be some real checks added here
        (_('Placeholder'), _check_placeholder),
        (_('Policy File JSON to YAML Migration'),
         (common_checks.check_policy_json, {'conf': cfg.CONF})),
    )


def main():
    return upgradecheck.main(
        cfg.CONF, project='manila', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())
